% Demonstration d'un filtre a particule pour un monde rectangulaire.
% % (c) 2010-2011 Philippe Giguere GLO-7021/GLO-4001
% Indique l'ambiguite qui ne peut pas etre resolue.
% Le robot possede deux lasers : un devant, et un a gauche.
% Il ne possede pas de compas.
clearvars;
clf;
myFontSize = 16;

% Carte du monde, connue du robot, modelisee par des segments de droite.
Monde = [0 3  3  10 10 0 0; ...
         1.5 1.5  0  0   3 3 1.5];

% Variables CACHEES dcrivant le systme. Le filtre ne connaitra jamais
% ces valeurs. On doit les connaitre dans matlab pour simuler le
% deplacement veritable du robot.
SBeacon  = 0.01;       % Erreur en position pour la distance du repere
Somega   = 0.0;        % Bruit veritable sur la vitesse angulaire rad/s
SV       = 0.02;       % Bruit veritable sur la vitesse lineaire, en m/s
xVrai    = [5 0.5 0]'; % Position initiale du robot

% A partir de quel fraction on fait du resampling. Neff = RatioResampling*nParticules
RatioResampling = 0.4;

% Parametre de deplacement du robot
V = 1.0;     % Vitesse lineaire, en m/s
omega = 0.0; % Vitesse angulaire, en rad/s

SVEstime = 1.0;     % Valeur du bruit sur la vitesse lineaire pour deplacement des particules
SOmegaEstime = 0.3; % Valeur du bruit sur la vitesse angulaire pour deplacement des particules
SlaserEstime = 0.5; % Valeur du bruit sur laser estime dans le filtre particules

% Parametre d'un scanneur laser
LRF.StartAngle  = -0;   % Angle de depart
LRF.dAngle      =  90;  % Angle entre chaque rayon
LRF.EndAngle    =  90;  % Angle de fing
LRF.Draw        = true; % Tracer les lignes rouges montrant les rayons
LRF.noise       = 0.1;  % Valeur relle du bruit sur le capteur laser
LRF.LaserRange  = 20.0; % Portee du laser

% Cet autre scanneur laser est pour simuler un scan pour chaque particule
LRFParait = LRF;          
LRFParait.noise = 0;
LRFParait.Draw  = false;

nStep = 80;         % Nombre de pas dans la simulation
dT = 0.05;          % Increment de temps
nParticules = 1000; % Nombre de particules utilisees

% Je vais initialiser au hasard les particules, dans une boite de 10m x 3m,
% car je ne sais pas ou le robot se trouve.
X(1,:) = 10*rand(1,nParticules); % position en x
X(2,:) = 3*rand(1,nParticules);  % position en y
X(3,:) = 0.5*pi*floor(4*rand(1,nParticules)); % orientation
w = ones(1,nParticules)/nParticules;  % Poids egaux pour les particules

C = [-0.1 10 0 3.1]; % Les axes utilises

% Montrer la distribution des particules au depart
clf
PlotParticules2D(X,w,'o',0.1);
plot(Monde(1,:),Monde(2,:),'k-','LineWidth',3);
title('Valeur initiale des particules');
axis(C);
drawnow();
hold on;


% ============== Boucle du filtre a particules =================
for iStep = 1:nStep
    clf

    % ======= Debut de la simulation du systeme reel =============
    time = iStep*dT;

    % Je simule la vraie position du robot. Ces valeurs sont inconnues du
    % filtre a particules et du robot.
    Vsimul = V+SV*randn;
    xVrai(1) = xVrai(1) + Vsimul*cos(xVrai(3))*dT;
    xVrai(2) = xVrai(2) + Vsimul*sin(xVrai(3))*dT;
    xVrai(3) = xVrai(3) + (omega+Somega*randn)*dT;
    
    % Le laser prend les deux mesures. Ce sont les mesures veritables
    [ScanLineEnd]   = GetLaserScanPolygon(xVrai(1:2),xVrai(3)*57.3,Monde,LRF);
    % ======= Fin de la simulation du systeme reel =============
    
    % ================ Debut du filtre a particules =============
    for iParticule = 1:nParticules
        % Je propage chaque particule
        Vused = V+SVEstime*randn;
        X(1,iParticule) = X(1,iParticule) + (Vused)*cos(X(3,iParticule))*dT;
        X(2,iParticule) = X(2,iParticule) + (Vused)*sin(X(3,iParticule))*dT;
        X(3,iParticule) = X(3,iParticule) + (omega+SOmegaEstime*randn)*dT;
        
        % Mise-a-jour des poids
        [ScanLineParticle]   = GetLaserScanPolygon(X(1:2,iParticule),X(3,iParticule)*57.3,Monde,LRFParait);
        
        for index = 1:size(ScanLineParticle,2)
            wnew = gauss(ScanLineParticle(2,index)-ScanLineEnd(2,index),SlaserEstime);
            w(iParticule) = w(iParticule).*wnew;
        end

    end
    [X w] = ParticuleResampling(X,w,RatioResampling);
    % ================ Fin du filtre a particules =============
    
    % Montrer la position et le poids des particules
    PlotParticules2D(X,w,'o',0.1);
    hold on;
    plot(Monde(1,:),Monde(2,:),'k-','LineWidth',3);
    
    % Position moyenne ponderee. Important! Cette position ne fait de sens que si
    % la distribution est unimodale. Dans l'exemple ici, cette valeur
    % fait peu de sens, car on a une distribution bimodale.
    Xmean = sum(X(1,:).*w);
    Ymean = sum(X(2,:).*w);
    Hmean = sum(X(3,:).*w);
    plot(Xmean,Ymean,'go','MarkerFaceColor','g','MarkerSize',7);
    l = line([Xmean  Xmean+0.2*cos(Hmean)],[Ymean Ymean+0.2*sin(Hmean)]);
    set(l,'Color','g','LineWidth',3);
    
    % Montrer la position reelle du robot
    plot(xVrai(1),xVrai(2),'gs','MarkerSize',15);
    
    title(sprintf('Temps = %.2f s',time),'FontSize',myFontSize);
    xlabel('Position x','FontSize',myFontSize);
    ylabel('Position y','FontSize',myFontSize);
    set(gca,'FontSize',myFontSize-2);
    axis equal
    axis(C);
    drawnow;
end

